
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_OCTREE_HH_
#define _MOOF_OCTREE_HH_

#include <list>

#include <boost/shared_ptr.hpp>

#include <stlplus/ntree.hpp>

#include <Moof/Aabb.hh>
#include <Moof/Drawable.hh>
#include <Moof/Entity.hh>
#include <Moof/Math.hh>
#include <Moof/Sphere.hh>
	

namespace Mf {


struct OctreeNode : public Entity
{
	std::list<EntityPtr> objects;

	OctreeNode()
	{
		aabb_.min = Vector3(-1.0, -1.0, -1.0);
		aabb_.max = Vector3(1.0, 1.0, 1.0);
		sphere_.init(Vector3(0.0, 0.0, 0.0), 1.41421);
	}

	OctreeNode(const Aabb& aabb)
	{
		aabb_ = aabb;
		sphere_.point = aabb.getCenter();
		sphere_.radius = (aabb.min - sphere_.point).length();
	}

	void draw(Scalar alpha) const
	{
		std::list<EntityPtr>::const_iterator it;

		for (it = objects.begin(); it != objects.end(); ++it)
		{
			(*it)->draw(alpha);
		}
		if (!objects.empty())
			aabb_.draw(); // temporary
	}

	void drawIfVisible(Scalar alpha, const Camera& cam) const
	{
		std::list<EntityPtr>::const_iterator it;

		for (it = objects.begin(); it != objects.end(); ++it)
		{
			(*it)->drawIfVisible(alpha, cam);
		}
		if (!objects.empty())
			aabb_.draw();
	}


	bool isVisible(const Camera& cam) const
	{
		if (sphere_.isVisible(cam))
		{
			return aabb_.isVisible(cam);
		}

		return false;
	}
};


class Octree;
typedef boost::shared_ptr<Octree> OctreePtr;

class Octree
{

	stlplus::ntree<OctreeNode> root_;

public:

	explicit Octree(const OctreeNode& rootNode)
	{
		root_.insert(rootNode);
	}

	void insert(EntityPtr entity)
	{
		insert(root_.root(), entity);
	}

	void insert(stlplus::ntree<OctreeNode>::iterator node, EntityPtr entity)
	{
		Plane::Halfspace halfspace;
		int octantNum = -1;

		if (!node.valid())
		{
			std::cerr << "cannot insert into invalid node" << std::endl;
			return;
		}

		Plane xy = node->getAabb().getPlaneXY();
		halfspace = xy.intersectsSphere(entity->getSphere());

		if (halfspace == Plane::POSITIVE)
		{
			Plane xz = node->getAabb().getPlaneXZ();
			halfspace = xz.intersectsSphere(entity->getSphere());

			if (halfspace == Plane::POSITIVE)
			{
				Plane yz = node->getAabb().getPlaneYZ();
				halfspace = yz.intersectsSphere(entity->getSphere());

				if (halfspace == Plane::POSITIVE)
				{
					octantNum = 2;
				}
				else if (halfspace == Plane::NEGATIVE)
				{
					octantNum = 3;
				}
			}
			else if (halfspace == Plane::NEGATIVE)
			{
				Plane yz = node->getAabb().getPlaneYZ();
				halfspace = yz.intersectsSphere(entity->getSphere());

				if (halfspace == Plane::POSITIVE)
				{
					octantNum = 1;
				}
				else if (halfspace == Plane::NEGATIVE)
				{
					octantNum = 0;
				}
			}
		}
		else if (halfspace == Plane::NEGATIVE)
		{
			Plane xz = node->getAabb().getPlaneXZ();
			halfspace = xz.intersectsSphere(entity->getSphere());

			if (halfspace == Plane::POSITIVE)
			{
				Plane yz = node->getAabb().getPlaneYZ();
				halfspace = yz.intersectsSphere(entity->getSphere());

				if (halfspace == Plane::POSITIVE)
				{
					octantNum = 6;
				}
				else if (halfspace == Plane::NEGATIVE)
				{
					octantNum = 7;
				}
			}
			else if (halfspace == Plane::NEGATIVE)
			{
				Plane yz = node->getAabb().getPlaneYZ();
				halfspace = yz.intersectsSphere(entity->getSphere());

				if (halfspace == Plane::POSITIVE)
				{
					octantNum = 5;
				}
				else if (halfspace == Plane::NEGATIVE)
				{
					octantNum = 4;
				}
			}
		}

		if (octantNum == -1)
		{
			node->objects.push_front(entity);
			//return node;
		}
		else
		{
			if (root_.children(node) == 0)
			{
				addChildren(node);
			}

			stlplus::ntree<OctreeNode>::iterator child = root_.child(node, octantNum);

			if (child.valid())
			{
				return insert(child, entity);
			}
			else
			{
				std::cerr << "expected but found no child at index " << octantNum << std::endl;
				//return stlplus::ntree<OctreeNode>::iterator();
			}
			//return WeakPtr();
		}
	}

	void addChildren(stlplus::ntree<OctreeNode>::iterator node)
	{
		Aabb octant;

		if (!node.valid())
		{
			std::cerr << "cannot add children to invalid node" << std::endl;
			return;
		}

		for (int i = 0; i < 8; ++i)
		{
			node->getAabb().getOctant(octant, i);
			//OctreeNode octantNode(octant);

			root_.append(node, octant);
		}
	}

	void draw(stlplus::ntree<OctreeNode>::iterator node, Scalar alpha)
	{
		if (!node.valid())
		{
			std::cerr << "cannot draw null child node :-(" << std::endl;
			return;
		}

		node->draw(alpha);

		for (unsigned i = 0; i < root_.children(node); ++i)
		{
			stlplus::ntree<OctreeNode>::iterator child = root_.child(node, i);

			if (child.valid())
			{
				draw(child, alpha);
			}
			else
			{
				std::cerr << "node is not a leaf, but has an invalid child" << std::endl;
			}

		}
	}

	void drawIfVisible(stlplus::ntree<OctreeNode>::iterator node,
			Scalar alpha, const Camera& cam)
	{
		//node.drawIfVisible(alpha, cam);
		
		if (!node.valid())
		{
			std::cerr << "invalid child while drawing :-(" << std::endl;
			return;
		}

		Frustum::Collision collision =
			cam.getFrustum().containsSphere(node->getSphere());
		if (collision == Frustum::OUTSIDE) return;

		collision = cam.getFrustum().containsAabb(node->getAabb());
		if (collision == Frustum::OUTSIDE) return;


		if (collision == Frustum::INSIDE)
		{
			node->draw(alpha);
		}
		else // collision == Frustum::INTERSECT
		{
			node->drawIfVisible(alpha, cam);
		}

		if (root_.children(node) > 0)
		{
			if (collision == Frustum::INSIDE)
			{
				for (unsigned i = 0; i < root_.children(node); ++i)
				{
					stlplus::ntree<OctreeNode>::iterator child = root_.child(node, i);

					if (child.valid())
					{
						draw(child, alpha);
					}
					else
					{
						std::cerr << "node is not a leaf, but has an invalid child" << std::endl;
					}

				}
			}
			else // collision == Frustum::INTERSECT
			{
				for (unsigned i = 0; i < root_.children(node); ++i)
				{
					stlplus::ntree<OctreeNode>::iterator child = root_.child(node, i);

					if (child.valid())
					{
						drawIfVisible(child, alpha, cam);
					}
					else
					{
						std::cerr << "node is not a leaf, but has an invalid child" << std::endl;
					}
				}
			}
		}
	}

	void drawIfVisible(Scalar alpha, const Camera& cam)
	{
		drawIfVisible(root_.root(), alpha, cam);
	}

};



} // namespace Mf

#endif // _MOOF_OCTREE_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

